from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from typing import Any

from django.conf import settings
from django.http import HttpRequest
from qstash import Receiver

from django_qstash.db.models import TaskStatus

from . import utils
from .exceptions import PayloadError
from .exceptions import SignatureError
from .exceptions import TaskError
from .results.services import store_task_result

logger = logging.getLogger(__name__)


@dataclass
class TaskPayload:
    function: str
    module: str
    args: list
    kwargs: dict
    task_name: str
    function_path: str

    @classmethod
    def from_dict(cls, data: dict) -> TaskPayload:
        """Create TaskPayload from dictionary."""
        is_valid, error = utils.validate_task_payload(data)
        if not is_valid:
            raise PayloadError(error)

        function_path = f"{data['module']}.{data['function']}"
        return cls(
            function=data["function"],
            module=data["module"],
            args=data["args"],
            kwargs=data["kwargs"],
            task_name=data.get("task_name", function_path),
            function_path=function_path,
        )


class QStashWebhook:
    def __init__(self):
        self.receiver = Receiver(
            current_signing_key=settings.QSTASH_CURRENT_SIGNING_KEY,
            next_signing_key=settings.QSTASH_NEXT_SIGNING_KEY,
        )
        self.force_https = getattr(settings, "DJANGO_QSTASH_FORCE_HTTPS", True)

    def verify_signature(self, body: str, signature: str, url: str) -> None:
        """Verify QStash signature."""
        if not signature:
            raise SignatureError("Missing Upstash-Signature header")

        if self.force_https and not url.startswith("https://"):
            url = url.replace("http://", "https://")

        try:
            self.receiver.verify(body=body, signature=signature, url=url)
        except Exception as e:
            raise SignatureError(f"Invalid signature: {e}")

    def parse_payload(self, body: str) -> TaskPayload:
        """Parse and validate webhook payload."""
        try:
            data = json.loads(body)
            return TaskPayload.from_dict(data)
        except json.JSONDecodeError as e:
            raise PayloadError(f"Invalid JSON payload: {e}")

    def execute_task(self, payload: TaskPayload) -> Any:
        """Import and execute the task function."""
        try:
            task_func = utils.import_string(payload.function_path)
        except ImportError as e:
            raise TaskError(f"Could not import task function: {e}")

        try:
            if callable(task_func) and hasattr(task_func, "actual_func"):
                return task_func.actual_func(*payload.args, **payload.kwargs)
            return task_func(*payload.args, **payload.kwargs)
        except Exception as e:
            raise TaskError(f"Task execution failed: {e}")

    def handle_request(self, request: HttpRequest) -> tuple[dict, int]:
        """Process webhook request and return response data and status code."""
        payload = None
        task_id = request.headers.get("Upstash-Message-Id")

        try:
            body = request.body.decode()
            self.verify_signature(
                body=body,
                signature=request.headers.get("Upstash-Signature"),
                url=request.build_absolute_uri(),
            )

            payload = self.parse_payload(body)
            result = self.execute_task(payload)
            store_task_result(
                task_id=task_id,
                task_name=payload.task_name,
                status=TaskStatus.SUCCESS,
                result=result,
                args=payload.args,
                kwargs=payload.kwargs,
                function_path=payload.function_path,
            )

            return {
                "status": "success",
                "task_name": payload.task_name,
                "result": result if result is not None else "null",
            }, 200

        except (SignatureError, PayloadError) as e:
            logger.exception("Authentication error: %s", str(e))
            return {
                "status": "error",
                "error_type": e.__class__.__name__,
                "error": str(e),
                "task_name": getattr(payload, "task_name", None),
            }, 400

        except TaskError as e:
            logger.exception("Task execution error: %s", str(e))
            store_task_result(
                task_id=task_id,
                task_name=payload.task_name,
                status=TaskStatus.EXECUTION_ERROR,
                traceback=str(e),
                args=payload.args,
                kwargs=payload.kwargs,
                function_path=payload.function_path,
            )
            return {
                "status": "error",
                "error_type": e.__class__.__name__,
                "error": str(e),
                "task_name": payload.task_name,
            }, 422

        except Exception as e:
            logger.exception("Unexpected error in webhook handler: %s", str(e))
            if payload:  # Store unexpected errors only if payload was parsed
                store_task_result(
                    task_id=task_id,
                    task_name=payload.task_name,
                    status=TaskStatus.INTERNAL_ERROR,
                    traceback=str(e),
                    args=payload.args,
                    kwargs=payload.kwargs,
                    function_path=payload.function_path,
                )
            return {
                "status": "error",
                "error_type": "InternalServerError",
                "error": "An unexpected error occurred",
                "task_name": getattr(payload, "task_name", None),
            }, 500



from __future__ import annotations

import pytest

from django_qstash.exceptions import PayloadError
from django_qstash.exceptions import SignatureError
from django_qstash.exceptions import TaskError
from django_qstash.exceptions import WebhookError


def test_webhook_error():
    """Test that WebhookError can be raised and is an Exception"""
    with pytest.raises(WebhookError):
        raise WebhookError("Test webhook error")


def test_signature_error():
    """Test that SignatureError can be raised and is a WebhookError"""
    with pytest.raises(SignatureError):
        raise SignatureError("Invalid signature")

    # Verify inheritance
    assert issubclass(SignatureError, WebhookError)


def test_payload_error():
    """Test that PayloadError can be raised and is a WebhookError"""
    with pytest.raises(PayloadError):
        raise PayloadError("Invalid payload")

    # Verify inheritance
    assert issubclass(PayloadError, WebhookError)


def test_task_error():
    """Test that TaskError can be raised and is a WebhookError"""
    with pytest.raises(TaskError):
        raise TaskError("Task execution failed")

    # Verify inheritance
    assert issubclass(TaskError, WebhookError)


def test_error_messages():
    """Test that error messages are properly stored"""
    message = "Custom error message"

    try:
        raise WebhookError(message)
    except WebhookError as e:
        assert str(e) == message

    try:
        raise SignatureError(message)
    except SignatureError as e:
        assert str(e) == message



from __future__ import annotations

import warnings

from django.conf import settings

QSTASH_TOKEN = getattr(settings, "QSTASH_TOKEN", None)
DJANGO_QSTASH_DOMAIN = getattr(settings, "DJANGO_QSTASH_DOMAIN", None)
DJANGO_QSTASH_WEBHOOK_PATH = getattr(
    settings, "DJANGO_QSTASH_WEBHOOK_PATH", "/qstash/webhook/"
)
if not QSTASH_TOKEN or not DJANGO_QSTASH_DOMAIN:
    warnings.warn(
        "DJANGO_SETTINGS_MODULE (settings.py required) requires QSTASH_TOKEN and DJANGO_QSTASH_DOMAIN should be set for QStash functionality",
        RuntimeWarning,
        stacklevel=2,
    )



from __future__ import annotations

import logging
import os
import warnings
from functools import lru_cache
from importlib import import_module

from django.apps import apps
from django.conf import settings
from django.core.signals import request_started
from django.utils.module_loading import module_has_submodule

logger = logging.getLogger(__name__)

DJANGO_QSTASH_DISCOVER_INCLUDE_SETTINGS_DIR = getattr(
    settings, "DJANGO_QSTASH_DISCOVER_INCLUDE_SETTINGS_DIR", True
)


@lru_cache(maxsize=None)
def discover_tasks(locations_only: bool = False) -> list[str] | list[dict]:
    """
    Automatically discover tasks in Django apps and return them as a list of tuples.
    Each tuple contains (dot_notation_path, task_name).
    If no custom task name is specified, both values will be the dot notation path.

    Returns:
        List of tuples: [(dot_notation_path, task_name), ...]
        Example: [
            ('example_app.tasks.my_task', 'example_app.tasks.my_task'),
            ('other_app.tasks.custom_task', 'special_name')
        ]
    """
    from django_qstash.app import QStashTask

    discovered_tasks = []
    packages = []

    # Add Django apps that contain tasks.py
    for app_config in apps.get_app_configs():
        if module_has_submodule(app_config.module, "tasks"):
            packages.append(app_config.name)

    # Add the directory containing settings.py if it has a tasks.py module
    if DJANGO_QSTASH_DISCOVER_INCLUDE_SETTINGS_DIR:
        settings_module = os.environ.get("DJANGO_SETTINGS_MODULE", "")
        if settings_module:
            settings_package = settings_module.rsplit(".", 1)[0]
            try:
                settings_module_obj = import_module(settings_package)
                if module_has_submodule(settings_module_obj, "tasks"):
                    packages.append(settings_package)
            except ImportError:
                warnings.warn(
                    f"Could not import settings package {settings_package} for task discovery",
                    RuntimeWarning,
                    stacklevel=2,
                )

    # Rest of the discovery logic
    for package in packages:
        try:
            tasks_module = import_module(f"{package}.tasks")
            # Find all attributes that are QstashTask instances
            for attr_name in dir(tasks_module):
                attr = getattr(tasks_module, attr_name)

                if isinstance(attr, QStashTask):
                    value = f"{package}.tasks.{attr_name}"
                    if attr.name == attr_name:
                        label = value
                    else:
                        label = f"{attr.name} ({package}.tasks)"
                    discovered_tasks.append(
                        {
                            "name": attr.name,
                            "field_label": label,
                            "location": f"{package}.tasks.{attr_name}",
                        }
                    )
        except Exception as e:
            warnings.warn(
                f"Failed to import tasks from {package}: {str(e)}",
                RuntimeWarning,
                stacklevel=2,
            )
    if locations_only:
        return [x["location"] for x in discovered_tasks]
    return discovered_tasks


def clear_discover_tasks_cache(sender, **kwargs):
    logger.info("Clearing Django QStash discovered tasks cache")
    discover_tasks.cache_clear()


request_started.connect(
    clear_discover_tasks_cache,
    dispatch_uid="clear_django_qstash_discovered_tasks_cache",
)



from __future__ import annotations

import json
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from django.http import HttpRequest

from django_qstash.exceptions import PayloadError
from django_qstash.exceptions import SignatureError
from django_qstash.exceptions import TaskError
from django_qstash.handlers import QStashWebhook
from django_qstash.handlers import TaskPayload

# Add pytest mark for database access
pytestmark = pytest.mark.django_db


class TestTaskPayload:
    def test_from_dict_valid(self):
        data = {
            "function": "test_func",
            "module": "test_module",
            "args": [1, 2],
            "kwargs": {"key": "value"},
        }
        payload = TaskPayload.from_dict(data)

        assert payload.function == "test_func"
        assert payload.module == "test_module"
        assert payload.args == [1, 2]
        assert payload.kwargs == {"key": "value"}
        assert payload.task_name == "test_module.test_func"
        assert payload.function_path == "test_module.test_func"

    def test_from_dict_invalid(self):
        data = {"invalid": "data"}
        with pytest.raises(PayloadError):
            TaskPayload.from_dict(data)


class TestQStashWebhook:
    @pytest.fixture
    def webhook(self):
        return QStashWebhook()

    def test_verify_signature_missing(self, webhook):
        with pytest.raises(SignatureError, match="Missing Upstash-Signature"):
            webhook.verify_signature("body", "", "http://example.com")

    def test_verify_signature_invalid(self, webhook):
        with pytest.raises(SignatureError, match="Invalid signature"):
            webhook.verify_signature("body", "invalid", "http://example.com")

    def test_parse_payload_invalid_json(self, webhook):
        with pytest.raises(PayloadError, match="Invalid JSON payload"):
            webhook.parse_payload("invalid json")

    def test_execute_task_import_error(self, webhook):
        payload = Mock(function_path="nonexistent.module")
        with pytest.raises(TaskError, match="Could not import task function"):
            webhook.execute_task(payload)

    def test_handle_request_success(self, webhook):
        request = Mock(spec=HttpRequest)
        request.body = json.dumps(
            {"function": "test_func", "module": "test_module", "args": [], "kwargs": {}}
        ).encode()
        request.headers = {"Upstash-Signature": "valid", "Upstash-Message-Id": "123"}
        request.build_absolute_uri.return_value = "https://example.com"

        with (
            patch.object(webhook, "verify_signature"),
            patch.object(webhook, "execute_task") as mock_execute,
        ):
            mock_execute.return_value = "result"
            response, status = webhook.handle_request(request)

        assert status == 200
        assert response["status"] == "success"
        assert response["result"] == "result"

    def test_handle_request_signature_error(self, webhook):
        request = Mock(spec=HttpRequest)
        request.body = json.dumps(
            {"function": "test_func", "module": "test_module", "args": [], "kwargs": {}}
        ).encode()
        request.headers = {}
        request.build_absolute_uri.return_value = "https://example.com"

        response, status = webhook.handle_request(request)

        assert status == 400
        assert response["status"] == "error"
        assert response["error_type"] == "SignatureError"

    def test_handle_request_task_error(self, webhook):
        request = Mock(spec=HttpRequest)
        request.body = json.dumps(
            {"function": "test_func", "module": "test_module", "args": [], "kwargs": {}}
        ).encode()
        request.headers = {"Upstash-Signature": "valid", "Upstash-Message-Id": "123"}
        request.build_absolute_uri.return_value = "https://example.com"

        with (
            patch.object(webhook, "verify_signature"),
            patch.object(webhook, "execute_task") as mock_execute,
        ):
            mock_execute.side_effect = TaskError("Task failed")
            response, status = webhook.handle_request(request)

        assert status == 422
        assert response["status"] == "error"
        assert response["error_type"] == "TaskError"
        assert response["error"] == "Task failed"
        assert response["task_name"] is not None

    def test_handle_request_unexpected_error(self, webhook):
        request = Mock(spec=HttpRequest)
        request.body = json.dumps(
            {"function": "test_func", "module": "test_module", "args": [], "kwargs": {}}
        ).encode()
        request.headers = {"Upstash-Signature": "valid"}
        request.build_absolute_uri.return_value = "https://example.com"

        with patch.object(webhook, "verify_signature") as mock_verify:
            mock_verify.side_effect = Exception("Unexpected error")
            response, status = webhook.handle_request(request)

        assert status == 500
        assert response["status"] == "error"
        assert response["error_type"] == "InternalServerError"
        assert response["error"] == "An unexpected error occurred"
        assert response["task_name"] is None

    def test_execute_task_with_actual_func(self, webhook):
        def actual_function(*args, **kwargs):
            return "actual result"

        mock_func = Mock()
        mock_func.actual_func = actual_function

        payload = Mock(function_path="test.path", args=[1, 2], kwargs={"key": "value"})

        with patch(
            "django_qstash.handlers.utils.import_string", return_value=mock_func
        ):
            result = webhook.execute_task(payload)

        assert result == "actual result"



from __future__ import annotations

import uuid
from datetime import datetime

import pytest
from django.utils import timezone

from django_qstash.results.models import TaskResult


@pytest.mark.django_db(transaction=True)
class TestTaskResult:
    def test_create_task_result(self):
        task_result = TaskResult.objects.create(
            task_id="test-task-123",
            task_name="test_module.test_function",
        )

        assert isinstance(task_result.id, uuid.UUID)
        assert task_result.task_id == "test-task-123"
        assert task_result.task_name == "test_module.test_function"
        assert task_result.status == "PENDING"
        assert isinstance(task_result.date_created, datetime)
        assert task_result.date_done is None
        assert task_result.result is None
        assert task_result.traceback is None
        assert task_result.args is None
        assert task_result.kwargs is None

    def test_str_representation(self):
        task_result = TaskResult.objects.create(
            task_id="test-task-123",
            task_name="test_module.test_function",
        )

        assert str(task_result) == "test_module.test_function (test-task-123)"

    def test_task_result_with_all_fields(self):
        task_result = TaskResult.objects.create(
            task_id="test-task-456",
            task_name="test_module.other_function",
            status="SUCCESS",
            date_done=timezone.now(),
            result={"success": True},
            traceback="No errors",
            args=[1, 2, 3],
            kwargs={"key": "value"},
        )

        assert task_result.status == "SUCCESS"
        assert isinstance(task_result.date_done, datetime)
        assert task_result.result == {"success": True}
        assert task_result.traceback == "No errors"
        assert task_result.args == [1, 2, 3]
        assert task_result.kwargs == {"key": "value"}

    def test_ordering(self):
        # Create tasks with different completion dates
        older_task = TaskResult.objects.create(
            task_id="older-task",
            task_name="test.older",
            status="SUCCESS",
            date_done=timezone.now(),
        )
        newer_task = TaskResult.objects.create(
            task_id="newer-task",
            task_name="test.newer",
            status="SUCCESS",
            date_done=timezone.now(),
        )

        # Get all tasks and verify ordering
        tasks = TaskResult.objects.all()
        assert tasks[0].task_id == newer_task.task_id
        assert tasks[1].task_id == older_task.task_id



import pytest

from django_qstash.utils import import_string, validate_task_payload


def test_import_string():
    """Test importing functions by string path"""
    # Import a real Python function
    func = import_string("json.dumps")
    assert callable(func)

    # Test invalid import
    with pytest.raises(ImportError):
        import_string("nonexistent.module.function")


def test_validate_task_payload():
    """Test task payload validation"""
    # Valid payload
    valid_payload = {
        "function": "task_func",
        "module": "my_app.tasks",
        "args": [1, 2, 3],
        "kwargs": {"key": "value"},
    }
    is_valid, message = validate_task_payload(valid_payload)
    assert is_valid
    assert message == ""

    # Missing required field
    invalid_payload = {
        "function": "task_func",
        "args": [1, 2, 3],
        "kwargs": {"key": "value"},
    }
    is_valid, message = validate_task_payload(invalid_payload)
    assert not is_valid
    assert "module" in message

    # Invalid args type
    invalid_payload = {
        "function": "task_func",
        "module": "my_app.tasks",
        "args": "not a list",
        "kwargs": {"key": "value"},
    }
    is_valid, message = validate_task_payload(invalid_payload)
    assert not is_valid
    assert "Args must be" in message



from __future__ import annotations

import os
import warnings

from qstash import QStash

from django_qstash.settings import QSTASH_TOKEN

QSTASH_URL = os.environ.get("QSTASH_URL", None)


def init_qstash():
    kwargs = {
        "token": QSTASH_TOKEN,
    }
    if QSTASH_URL is not None:
        warning_msg = f"\n\n\033[93mUsing {QSTASH_URL} as your QStash URL. \
            \nThis configuration should only be used in development.\n\033[0m"
        warnings.warn(warning_msg, RuntimeWarning, stacklevel=2)
        kwargs["base_url"] = QSTASH_URL
    return QStash(**kwargs)


qstash_client = init_qstash()



from __future__ import annotations

from .base import AsyncResult
from .base import QStashTask
from .decorators import shared_task
from .decorators import stashed_task

__all__ = ["AsyncResult", "QStashTask", "stashed_task", "shared_task"]



from __future__ import annotations

import json
from typing import Any

from django_qstash.callbacks import get_callback_url
from django_qstash.schedules.models import TaskSchedule


def prepare_qstash_payload(instance: TaskSchedule) -> dict[str, Any]:
    """Prepare the task payload for QStash"""
    return {
        "function": instance.task_name.split(".")[-1],  # Get function name
        "module": ".".join(instance.task_name.split(".")[:-1]),  # Get module path
        "args": instance.args,
        "kwargs": instance.kwargs,
        "task_name": instance.name,
        "options": {
            "max_retries": instance.retries,
            "timeout": instance.timeout,
        },
    }


def format_task_schedule_for_qstash(instance: TaskSchedule) -> dict[str, Any]:
    payload = prepare_qstash_payload(instance)
    callback_url = get_callback_url()
    data = {
        "destination": callback_url,
        "body": json.dumps(payload),
        "cron": instance.cron,
        "retries": instance.retries,
        "timeout": instance.timeout,
    }
    if instance.schedule_id:
        data["schedule_id"] = instance.schedule_id
    return data



# Generated by Django 5.1.4 on 2024-12-30 23:31

from __future__ import annotations

import uuid

import django.utils.timezone
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="TaskResult",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid1,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                (
                    "task_id",
                    models.CharField(db_index=True, max_length=255, unique=True),
                ),
                ("task_name", models.CharField(max_length=255)),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("PENDING", "Pending"),
                            ("SUCCESS", "Success"),
                            ("FAILURE", "Failure"),
                        ],
                        default="PENDING",
                        max_length=50,
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("date_done", models.DateTimeField(null=True)),
                ("result", models.JSONField(null=True)),
                ("traceback", models.TextField(null=True)),
                ("args", models.JSONField(null=True)),
                ("kwargs", models.JSONField(null=True)),
            ],
            options={
                "ordering": ["-date_done"],
            },
        ),
    ]



from __future__ import annotations

from django.db.models.signals import post_save
from django.db.models.signals import pre_delete
from django.dispatch import receiver

from django_qstash.schedules import services
from django_qstash.schedules.models import TaskSchedule


@receiver(post_save, sender=TaskSchedule)
def sync_schedule_to_qstash_receiver(sender, instance, created, **kwargs):
    """
    Sync the django-qstash TaskSchedule to QStash on save.
    """
    services.sync_task_schedule_instance_to_qstash(instance)


@receiver(pre_delete, sender=TaskSchedule)
def delete_schedule_from_qstash_receiver(sender, instance, **kwargs):
    services.delete_task_schedule_from_qstash(instance)



from __future__ import annotations

import logging

from django.db import transaction
from qstash.schedule import Schedule

from django_qstash.client import qstash_client
from django_qstash.schedules.formatters import format_task_schedule_for_qstash
from django_qstash.schedules.models import TaskSchedule

logger = logging.getLogger(__name__)


@transaction.atomic
def sync_task_schedule_instance_to_qstash(instance: TaskSchedule) -> TaskSchedule:
    """Sync a task schedule to QStash.

    Creates a new schedule if none exists, handles pause/resume state changes,
    and updates existing schedules when active.
    """
    data = format_task_schedule_for_qstash(instance)
    schedule_id = qstash_client.schedule.create(**data)
    if not instance.schedule_id:
        TaskSchedule.objects.filter(id=instance.id).update(schedule_id=schedule_id)

    sync_state_changes(instance)

    return instance


def sync_state_changes(instance: TaskSchedule) -> None:
    """Handle pause/resume state changes."""
    if instance.did_just_resume():
        try:
            qstash_client.schedule.resume(instance.schedule_id)
        except Exception:
            logger.exception("Failed to resume schedule %s", instance.schedule_id)
    elif instance.did_just_pause():
        try:
            qstash_client.schedule.pause(instance.schedule_id)
        except Exception:
            logger.exception("Failed to pause schedule %s", instance.schedule_id)


def get_task_schedule_from_qstash(
    instance: TaskSchedule, as_dict: bool = False
) -> Schedule | dict | None:
    """Get a schedule from QStash."""
    try:
        response = qstash_client.schedule.get(instance.schedule_id)
    except Exception:
        logger.exception("Failed to lookup schedule %s", instance.schedule_id)
        return None
    if response:
        return response.__dict__ if as_dict else response
    return None


def delete_task_schedule_from_qstash(instance: TaskSchedule) -> None:
    """Delete a schedule from QStash."""
    try:
        qstash_client.schedule.delete(instance.schedule_id)
    except Exception:
        logger.exception("Failed to delete schedule %s", instance.schedule_id)
    return



from __future__ import annotations

from django.conf import settings
from django.core.management.base import BaseCommand

from django_qstash.results.tasks import clear_stale_results_task

DJANGO_QSTASH_RESULT_TTL = getattr(settings, "DJANGO_QSTASH_RESULT_TTL", 604800)


class Command(BaseCommand):
    help = f"""Clears stale task results older than\n
    {DJANGO_QSTASH_RESULT_TTL} seconds (settings.DJANGO_QSTASH_RESULT_TTL)"""

    def add_arguments(self, parser):
        parser.add_argument(
            "--no-input",
            action="store_true",
            help="Do not ask for confirmation",
        )
        parser.add_argument(
            "--since",
            type=int,
            help="The number of seconds ago to clear results for",
        )
        parser.add_argument(
            "--delay", action="store_true", help="Offload request using django_qstash"
        )

    def handle(self, *args, **options):
        delay = options["delay"]
        no_input = options["no_input"]
        since = options.get("since") or DJANGO_QSTASH_RESULT_TTL
        user_confirm = not no_input
        if not delay:
            clear_stale_results_task(
                since=since, user_confirm=user_confirm, stdout=self.stdout
            )
        else:
            clear_stale_results_task.delay(
                since=since, user_confirm=user_confirm, stdout=self.stdout
            )



from __future__ import annotations

from datetime import timedelta

from django.conf import settings
from django.core.validators import MaxValueValidator
from django.db import models
from django.utils import timezone

from django_qstash.discovery.models import TaskField
from django_qstash.schedules.validators import validate_cron_expression
from django_qstash.schedules.validators import validate_duration_string

DJANGO_QSTASH_DOMAIN = getattr(settings, "DJANGO_QSTASH_DOMAIN", None)
DJANGO_QSTASH_WEBHOOK_PATH = getattr(
    settings, "DJANGO_QSTASH_WEBHOOK_PATH", "/qstash/webhook/"
)


class TaskSchedule(models.Model):
    """
    A model that represents a QStash Schedule for any given django-qstash Task.
    """

    schedule_id = models.CharField(
        max_length=255,
        unique=True,
        blank=True,
        null=True,
        verbose_name="Schedule ID",
        help_text="The schedule ID stored in QStash",
        db_index=True,
    )
    name = models.CharField(
        max_length=200,
        verbose_name="Name",
        help_text="Short Description For This Task Schedule",
    )
    task = TaskField()
    task_name = models.CharField(
        max_length=255, help_text="Original Python location of task", blank=True
    )
    args = models.JSONField(
        blank=True,
        default=list,
        verbose_name="Positional Arguments",
        help_text='JSON encoded positional arguments (Example: ["arg1", "arg2"])',
    )
    kwargs = models.JSONField(
        blank=True,
        default=dict,
        verbose_name="Keyword Arguments",
        help_text='JSON encoded keyword arguments (Example: {"argument": "value"})',
    )
    # Configured for Upstash Cron
    # https://upstash.com/docs/qstash/api/schedules/create#param-upstash-cron
    cron = models.CharField(
        max_length=255,
        default="*/5 * * * *",
        validators=[validate_cron_expression],
        verbose_name="Cron Expression",
        help_text="Cron expression for scheduling the task",
    )
    # Configured for Upstash Retries
    # https://upstash.com/docs/qstash/api/schedules/create#param-upstash-retries
    retries = models.IntegerField(
        default=3,
        verbose_name="Retries",
        validators=[MaxValueValidator(5)],
        help_text="Number of times to retry the task if it fails",
    )
    # Configured for Upstash Timeout
    # https://upstash.com/docs/qstash/api/schedules/create#param-upstash-timeout
    timeout = models.CharField(
        max_length=10,
        default="60s",
        verbose_name="Timeout",
        validators=[validate_duration_string],
        help_text="Duration string for task timeout (e.g., '1s', '5m', '2h'). "
        "See Max HTTP Connection Timeout on QStash pricing page for allowed values for your Upstash account.",
    )
    updated_at = models.DateTimeField(auto_now=True)
    is_active = models.BooleanField(default=True)
    active_at = models.DateTimeField(null=True, blank=True, default=timezone.now)
    is_paused = models.BooleanField(default=False)
    paused_at = models.DateTimeField(null=True, blank=True)
    is_resumed = models.BooleanField(default=False)
    resumed_at = models.DateTimeField(null=True, blank=True)

    def save(self, *args, **kwargs):
        current_task_name = self.task
        if not self.pk or self.task_name != current_task_name:
            self.task_name = current_task_name

        if not self.is_active:
            self.is_paused = True
            self.is_resumed = False
            self.paused_at = timezone.now()
            self.resumed_at = None
            self.active_at = None
            self.is_active = False
        elif self.is_active:
            self.is_paused = False
            self.is_resumed = True
            self.paused_at = None
            self.resumed_at = timezone.now()
            self.active_at = timezone.now()
        super().save(*args, **kwargs)

    def did_just_resume(self, delta_seconds: int = 60) -> bool:
        if not self.is_resumed or not self.resumed_at:
            return False
        now = timezone.now()
        delta_window = now - timedelta(seconds=delta_seconds)
        return self.resumed_at >= delta_window

    def did_just_pause(self, delta_seconds: int = 60) -> bool:
        if not self.is_paused or not self.paused_at:
            return False
        now = timezone.now()
        delta_window = now - timedelta(seconds=delta_seconds)
        return self.paused_at >= delta_window



from __future__ import annotations

import logging
from datetime import timedelta

from django.apps import apps
from django.conf import settings
from django.utils import timezone

from django_qstash import stashed_task
from django_qstash.db.models import TaskStatus

DJANGO_QSTASH_RESULT_TTL = getattr(settings, "DJANGO_QSTASH_RESULT_TTL", 604800)

logger = logging.getLogger(__name__)


@stashed_task(name="Cleanup Task Results")
def clear_stale_results_task(
    since=None, stdout=None, user_confirm=False, exclude_errors=True, *args, **options
):
    delta_seconds = since or DJANGO_QSTASH_RESULT_TTL
    cutoff_date = timezone.now() - timedelta(seconds=delta_seconds)
    TaskResult = None
    try:
        TaskResult = apps.get_model("django_qstash_results", "TaskResult")
    except LookupError as e:
        msg = "Django QStash Results not installed.\nAdd `django_qstash.results` to INSTALLED_APPS and run migrations."
        if stdout is not None:
            stdout.write(msg)
        logger.exception(msg)
        raise e
    qs_to_delete = TaskResult.objects.filter(date_done__lt=cutoff_date)
    if exclude_errors:
        qs_to_delete = qs_to_delete.exclude(
            status__in=[
                TaskStatus.EXECUTION_ERROR,
                TaskStatus.INTERNAL_ERROR,
                TaskStatus.OTHER_ERROR,
            ]
        )

    if user_confirm:
        user_input = input("Are you sure? (Y/n): ")
        if f"{user_input}".lower() != "y":
            msg = "Skipping deletion"
            if stdout is not None:
                stdout.write(msg)
            logger.info(msg)
            return

    if not qs_to_delete.exists():
        msg = "No stale Django QStash task results found"
        if stdout is not None:
            stdout.write(msg)
        else:
            logger.info(msg)
        return
    delete_msg = f"Deleting {qs_to_delete.count()} task results older than {cutoff_date} ({DJANGO_QSTASH_RESULT_TTL} seconds)"
    if stdout is not None:
        stdout.write(delete_msg)
    else:
        logger.info(delete_msg)
    try:
        deleted_count, _ = qs_to_delete.delete()
        msg = f"Successfully deleted {deleted_count} stale results."
        if stdout is not None:
            stdout.write(msg)
        else:
            logger.info(msg)
    except Exception as e:
        msg = f"Error deleting stale results: {e}"
        if stdout is not None:
            stdout.write(msg)
        logger.exception(msg)
        raise e


@stashed_task(name="Clear Task Error Results")
def clear_task_errors_task(
    since=None, stdout=None, user_confirm=False, *args, **options
):
    clear_stale_results_task(
        since=since, stdout=stdout, user_confirm=user_confirm, exclude_errors=False
    )



# Generated by Django 5.1.4 on 2025-01-02 06:28

from __future__ import annotations

import django.core.validators
import django.utils.timezone
from django.db import migrations
from django.db import models

import django_qstash.discovery.models
import django_qstash.schedules.validators


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="TaskSchedule",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "schedule_id",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        help_text="The schedule ID stored in QStash",
                        max_length=255,
                        null=True,
                        unique=True,
                        verbose_name="Schedule ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Short Description For This Task Schedule",
                        max_length=200,
                        verbose_name="Name",
                    ),
                ),
                ("task", django_qstash.discovery.models.TaskField(max_length=255)),
                (
                    "task_name",
                    models.CharField(
                        blank=True,
                        help_text="Original Python location of task",
                        max_length=255,
                    ),
                ),
                (
                    "args",
                    models.JSONField(
                        blank=True,
                        default=list,
                        help_text='JSON encoded positional arguments (Example: ["arg1", "arg2"])',
                        verbose_name="Positional Arguments",
                    ),
                ),
                (
                    "kwargs",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        help_text='JSON encoded keyword arguments (Example: {"argument": "value"})',
                        verbose_name="Keyword Arguments",
                    ),
                ),
                (
                    "cron",
                    models.CharField(
                        default="*/5 * * * *",
                        help_text="Cron expression for scheduling the task",
                        max_length=255,
                        verbose_name="Cron Expression",
                    ),
                ),
                (
                    "retries",
                    models.IntegerField(
                        default=3,
                        help_text="Number of times to retry the task if it fails",
                        validators=[django.core.validators.MaxValueValidator(5)],
                        verbose_name="Retries",
                    ),
                ),
                (
                    "timeout",
                    models.CharField(
                        default="60s",
                        help_text="Duration string for task timeout (e.g., '1s', '5m', '2h'). See Max HTTP Connection Timeout on QStash pricing page for allowed values for your Upstash account.",
                        max_length=10,
                        validators=[
                            django_qstash.schedules.validators.validate_duration_string
                        ],
                        verbose_name="Timeout",
                    ),
                ),
                ("is_active", models.BooleanField(default=True)),
                (
                    "active_at",
                    models.DateTimeField(
                        blank=True, default=django.utils.timezone.now, null=True
                    ),
                ),
                ("is_paused", models.BooleanField(default=False)),
                ("paused_at", models.DateTimeField(blank=True, null=True)),
                ("is_resumed", models.BooleanField(default=False)),
                ("resumed_at", models.DateTimeField(blank=True, null=True)),
            ],
        ),
    ]



import importlib
import logging
from typing import Any, Tuple

logger = logging.getLogger(__name__)


def import_string(import_path: str) -> Any:
    """
    Import a module path and return the attribute/class designated by the last name.

    Example:
        import_string('myapp.tasks.mytask') -> mytask function
    """
    try:
        module_path, class_name = import_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Could not import '{import_path}': {e}")


def validate_task_payload(payload: dict) -> Tuple[bool, str]:
    """Validate the task payload has all required fields"""
    required_fields = {"function", "module", "args", "kwargs"}
    missing_fields = required_fields - set(payload.keys())

    if missing_fields:
        return False, f"Missing required fields: {', '.join(missing_fields)}"

    if not isinstance(payload["args"], (list, tuple)):
        return False, "Args must be a list or tuple"

    if not isinstance(payload["kwargs"], dict):
        return False, "Kwargs must be a dictionary"

    return True, ""



from __future__ import annotations

import json

from django.http import HttpRequest
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

from .handlers import QStashWebhook


@csrf_exempt
@require_http_methods(["POST"])
def qstash_webhook_view(request: HttpRequest) -> HttpResponse:
    """Handle QStash webhook requests."""
    webhook = QStashWebhook()
    response_data, status_code = webhook.handle_request(request)
    return HttpResponse(
        json.dumps(response_data), status=status_code, content_type="application/json"
    )



from __future__ import annotations

from django import forms

from django_qstash.discovery.utils import discover_tasks
from django_qstash.discovery.validators import task_exists_validator


class TaskChoiceField(forms.ChoiceField):
    """
    A form field that provides choices from discovered QStash tasks
    """

    def __init__(self, *args, **kwargs):
        # Remove max_length if it's present since ChoiceField doesn't use it
        kwargs.pop("max_length", None)

        # Get tasks before calling parent to set choices
        tasks = discover_tasks(locations_only=False)

        # Convert tasks to choices using (task_name, task_name) format
        task_choices = [(task["location"], task["field_label"]) for task in tasks]

        kwargs["choices"] = task_choices
        kwargs["validators"] = [task_exists_validator] + kwargs.get("validators", [])
        super().__init__(*args, **kwargs)

    def get_task(self):
        """
        Returns the actual task dot notation path for the selected value
        """
        if self.data:
            tasks = discover_tasks(locations_only=False)

            for task in tasks:
                if task["field_label"] == self.data:
                    return task["location"]
        return None



from __future__ import annotations

from django import forms

from django_qstash.discovery.fields import TaskChoiceField
from django_qstash.schedules.models import TaskSchedule


class TaskScheduleForm(forms.ModelForm):
    task = TaskChoiceField()

    class Meta:
        model = TaskSchedule
        fields = [
            "name",
            "task",
            "task_name",
            "args",
            "kwargs",
            "schedule_id",
            "cron",
            "retries",
            "timeout",
        ]

    def clean(self):
        cleaned_data = super().clean()
        # If task_name is not provided, use the task value
        if not cleaned_data.get("task_name") and cleaned_data.get("task"):
            cleaned_data["task_name"] = cleaned_data["task"]
        return cleaned_data
